/*****************************************************************************
    dllapps.c

    Diese Datei ist Teil des Programmpakets 'ACL-tftp'.
    Alle Rechte bei Andreas G. Schindler
                    Klappacher Str. 2i
                    64285 Darmstadt, Deutschland.

    Dieses Programm ist 'Freie Software'; Sie knnen es entsprechend den 
    Bedingungen der 'Allgemeinen ffentlichen GNU-Lizenz' (verffentlicht 
    durch die 'Free Software Foundation') weiter verbreiten und/oder 
    modifizieren. Massgeblich ist die Version 2 oder eine neuere Version
    dieser Lizenz. 

    Dieses Programm wird in der Hoffnung dass es ntzlich is weitergegeben,
    jedoch OHNE JEDE GARANTIE. Dieser Garantieauschluss bezieht sich auch
    auf Eigenschaften wie Marktfhigkeit oder Eignung zu einem bestimmten
    Zweck. Details finden sich in der 'Allgemeinen ffentlichen GNU-Lizenz'.

    Eine Kopie dieser Lizenz sollte in  Programmpaket, dem diese Datei
    entstammt, enthalten sein. Falls nicht, schreiben Sie an die 
    'Free Software Foundation, Inc.', 59 Temple Place, Suite 330, Boston, 
    MA  02111-1307  USA.


    This file is part of the 'ACL-tftp' Package
    Copyright (C) 1999-2003 by Andreas G. Schindler
                               Klappacher Str. 2i
                               Darmstadt, Germany.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*****************************************************************************/
#include "auto.h"
/*
	This file contains specific functions depending on the dll-package
	that deal with regular expressions and ACLs.
*/ 

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include "globals.h"
#include "netio.h"
#include "access.h"
#include "util.h"

static void getHeadTail (void *ehead, void *headp, void *tailp)
{
    dll_t *p;

    p = (dll_t *)ehead;
    if (headp != NULL) *((dll_t **)headp) = p;
    if (tailp != NULL) *((dll_t **)tailp) = p+1;
}


void clearReChain (dlr_t *head)
{
    clearChain ((dll_t *)head);
}

void clearPidChain (dlp_t *head)
{
    clearChain ((dll_t *)head);
}

void clearFilterChain (dlf_t *head)
{
    clearChain ((dll_t *)head);
}

void deleteCacheChain (dlc_t *head)
{
    dlc_t *ptr, *tail; 

    getHeadTail (head, NULL, &tail);

    for (ptr = head->next; ptr != tail; ptr = ptr->next) {
        if (ptr->name != NULL)
            Free(ptr->name);
        if (ptr->cache != NULL)
            Free(ptr->cache);
    }

    deleteChain((dll_t *)head);
    currCachesize  = 0;
    currCachefiles = 0; 
}


void deleteReChain (dlr_t *head)
{
    dlr_t *ptr, *tail; 

    getHeadTail (head, NULL, &tail);

    for (ptr = head->next; ptr != tail; ptr = ptr->next) {
        regfree(&ptr->compiled);
        if (ptr->raw != NULL)
            Free(ptr->raw);
        if (ptr->map != NULL)
            Free(ptr->map);
    }

    deleteChain((dll_t *)head);
}


void deleteACLChain (dla_t *head)
{
    deleteChain((dll_t *)head);
}


void deleteTimerChain (dlt_t *head)
{
    deleteChain((dll_t *)head);
}


void deletePidChain (dlp_t *head)
{
    deleteChain((dll_t *)head);
}

void deleteFilterChain (dlf_t *head)
{
    dlf_t *ptr, *tail; 

    getHeadTail (head, NULL, &tail);

    for (ptr = head->next; ptr != tail; ptr = ptr->next) {
        if (ptr->rxfilter != NULL)
            Free(ptr->rxfilter);
        if (ptr->txfilter != NULL)
            Free(ptr->txfilter);
    }

    deleteChain((dll_t *)head);
}


void deleteBlistChain (dlb_t *head)
{
    deleteChain((dll_t *)head);
}

dlr_t *initReChain(void)
{
    return ((dlr_t *)initChain(NULL));
}


dla_t *initACLChain(void)
{
    return ((dla_t *)initChain(NULL));
}


dlt_t *initTimerChain(void)
{
    return ((dlt_t *)initChain(NULL));
}


dlp_t *initPidChain(void)
{
    return ((dlp_t *)initChain(NULL));
}

dlf_t *initFilterChain(void)
{
    return ((dlf_t *)initChain(NULL));
}

dlb_t *initBlistChain(void)
{
    return ((dlb_t *)initChain(NULL));
}

dlc_t *initCacheChain(void)
{
    return ((dlc_t *)initChain(NULL));
}

void deleteACLByMark (dla_t *head, int mark)
{
    deleteByMark ((dll_t *)head, mark, DLL_FREE);
}

void deleteFilterByMark (dlf_t *head, int mark)
{
    deleteByMark ((dll_t *)head, mark, DLL_FREE);
}


void deleteReByMark (dlr_t *head, int mark)
{
    dlr_t *ptr, *tail;

    getHeadTail (head, NULL, &tail);

    for (ptr = head->next; ptr != tail; ) {
        if (ptr->mark == mark) {
            regfree(&ptr->compiled);
            if (ptr->raw != NULL)
                Free(ptr->raw);
            if (ptr->map != NULL)
                Free(ptr->map);
            ptr = (dlr_t *)removeElement((dll_t *)ptr, DLL_GETNEXT|DLL_FREE, 
                                         (dll_t *)head);
        }
        else
            ptr = ptr->next;
    }
}


int deletePid (dlp_t *head, pid_t pid)
{
    dlp_t *ptr, *tail;

    getHeadTail (head, NULL, &tail);

    for (ptr = head->next; ptr != tail; ptr=ptr->next) {
        if (ptr->pid == pid) {
            removeElement((dll_t *)ptr, DLL_FREE, (dll_t *)head);
            return 0;
        }
    }
    return 1;
}


int deleteCacheEntry (dlc_t *head, dlc_t *entry)
{
    if (entry == 0)
        return 1;

    removeElement((dll_t *)entry, 0, (dll_t *)head);

    if (entry->name != NULL)
        Free (entry->name);

    if (entry->cache != NULL)
        Free (entry->cache);

    currCachefiles -= 1;
    currCachesize  -= entry->size;
    Free (entry);
    return 0;
}

dlr_t *findRePattern (dlr_t *head, dlr_t *start, int mark, 
                      char *name, int dtype)
{
    dlr_t *ptr, *tail;

    getHeadTail (head, NULL, &tail);

    if (start == NULL)
        start = head;
    for (ptr = start->next; ptr != tail; ptr = ptr->next) {
        if ((mark == 0) || (ptr->mark == mark))
            if ((dtype == 0) || (ptr->dtype & dtype))
                if (!regexec(&ptr->compiled, name, 0, NULL, 0)) 
                    return ptr;
    }
    return NULL;	
}

dla_t *findACL (dla_t *head, dla_t *start, int mark, in_addr_t addr, int mode)
{
    dla_t *ptr, *last=NULL, *tail;
    in_addr_t cmask;

    getHeadTail (head, NULL, &tail);

    if (start == NULL)
        start = head;

    if (mode == FIND_MIN)
        cmask = 0xffffffff;
    else
        cmask = 0;

    for (ptr = start->next; ptr != tail; ptr = ptr->next) {
        if (mark == 0 || ptr->mark == mark) {
            if ((addr & ptr->acl.mask) == (ptr->acl.addr & ptr->acl.mask)) {
                switch (mode) {
                  case FIND_LAST:
                    last = ptr;
                    break;
                  case FIND_MIN:
                    if (ptr->acl.mask < cmask) {
                        cmask = ptr->acl.mask;
                        last = ptr;
                    }
                    break;
                  case FIND_MAX:
                    if (ptr->acl.mask > cmask) {
                        cmask = ptr->acl.mask;
                        last = ptr;
                    }
                    break;
                  case FIND_FIRST:
                  default:
                    return ptr;
                    break;
                }
            }
        }
    }
    return last;
}


dlp_t *findPid (dlp_t *head, pid_t pid)
{
    dlp_t *ptr, *tail;

    getHeadTail (head, NULL, &tail);

    for (ptr = head->next; ptr != tail; ptr = ptr->next)
        if (ptr->pid == pid)
            return ptr;

    return NULL;
}


dlc_t *findCacheByMark (dlc_t *head, char *name, int mark)
{
    dlc_t *ptr, *tail;

    getHeadTail (head, NULL, &tail);

    for (ptr = head->next; ptr != tail; ptr = ptr->next)
        if (ptr->mark == mark)
            if (!strcmp(ptr->name, name))
                return ptr;

    return NULL;
}

dlr_t *getNextRe (dlr_t *head, dlr_t *start)
{
    return((dlr_t *)getNext((dll_t *)head, (dll_t *)start));
}

dlc_t *getNextCache (dlc_t *head, dlc_t *start)
{
    return((dlc_t *)getNext((dll_t *)head, (dll_t *)start));
}


dla_t *getNextACL (dla_t *head, dla_t *start)
{
    return((dla_t *)getNext((dll_t *)head, (dll_t *)start));
}


dlt_t *getNextTimer (dlt_t *head, dlt_t *start)
{
    return((dlt_t *)getNext((dll_t *)head, (dll_t *)start));
}


dlf_t *getNextFilter (dlf_t *head, dlf_t *start)
{
    return((dlf_t *)getNext((dll_t *)head, (dll_t *)start));
}


dlp_t *getNextPid (dlp_t *head, dlp_t *start)
{
    return((dlp_t *)getNext((dll_t *)head, (dll_t *)start));
}


dlb_t *getNextBlist (dlb_t *head, dlb_t *start)
{
    return((dlb_t *)getNext((dll_t *)head, (dll_t *)start));
}


int getPidCount (dlp_t *head)
{
    return (int)chainCount(head);
}

pid_t *getPidVec (dlp_t *head)
{
    dll_t *dp;
    pid_t *pp, *pidvec;

    if ((head == NULL) || !chainCount(head))
        return NULL;

    pp = pidvec = (pid_t *)Malloc((chainCount(head)+1)*sizeof(pid_t));

    if (pidvec == NULL)
        return NULL;

    *pp++ = (pid_t)chainCount(head);

    for (dp = NULL; (dp = getNext((dll_t *)head, dp)) != NULL;)
       *pp++ = ((dlp_t *)dp)->pid;

    return pidvec;
}

int getPidCountByMark (dlp_t *head, int mark)
{
    dll_t *dp;
    int count;

    for (dp = NULL, count = 0; (dp = getNext((dll_t *)head, dp)) != NULL;)
        if (dp->mark == mark)
            count++;

    return count;
}

dla_t *getACLByMark (dla_t *head, dla_t *start, int mark)
{
    return ((dla_t *)getByMark ((dll_t *)head, (dll_t *)start, mark));
}


dlt_t *getTimerByMark (dlt_t *head, dlt_t *start, int mark)
{
    return ((dlt_t *)getByMark ((dll_t *)head, (dll_t *)start, mark));
}


dlf_t *getFilterByMark (dlf_t *head, dlf_t *start, int mark)
{
    return ((dlf_t *)getByMark ((dll_t *)head, (dll_t *)start, mark));
}

dlr_t *getReByMark (dlr_t *head, dlr_t *start, int mark)
{
    return ((dlr_t *)getByMark ((dll_t *)head, (dll_t *)start, mark));
}

dlp_t *getPidByMark (dlp_t *head, dlp_t *start, int mark)
{
    return ((dlp_t *)getByMark ((dll_t *)head, (dll_t *)start, mark));
}

int addRePattern (dlr_t *head, int mark, 
                  char *pattern, char *map, int filter, int dtype, 
                  u_long maxsize, dlr_t **np)
{
    dlr_t *new;
    int rtc;
    int icase = 0, off = 0;

    if ((new = (dlr_t *)Malloc(sizeof(dlr_t))) == NULL)
        return -1;

    if (*pattern == '!') {	/* case insensitive flag */
        icase = REG_ICASE;
        off = 1;
    }

    if ((*pattern == '\\') && (*(pattern+1) == '!'))
        off = 1;

    if ((new->raw = Strdup(pattern)) == NULL) {
        Free(new);
        return -1;
    }

    if (map != NULL) {
        if ((new->map = Strdup(map)) == NULL) {
            Free(new->raw);
            Free(new);
            return -1;
        }
    }
    else {
        new->map = NULL;
    }
    
    if ((rtc = regcomp(&new->compiled, pattern+off, icase))) {
        Free(new->raw);
        Free(new->map);
        Free(new);
        return rtc;
    }

    new->ctype   = CTYPE_REGEX;
    new->mark    = mark;
    new->filter  = filter;
    new->dtype   = dtype;
    new->maxsize = maxsize;

    insertAtTail ((dll_t *)head, (dll_t *)new);
    if (np != NULL) *np = new;

    return 0;
}

int addCache (dlc_t *head, int mark, char *name, struct stat *status,
              int flags, dlc_t **np)
{
    dlc_t *new;

    if ((new = (dlc_t *)Malloc(sizeof(dlc_t))) == NULL)
        return -1;

    if ((new->name = Strdup(name)) == NULL) {
        Free(new);
        return -1;
    }

    new->size    = (ulong)status->st_size + 2;

    if ((new->cache = (char *)Malloc(new->size)) == NULL) {
        Free(new->name);
        Free(new);
        return -1;
    }

    new->ctype   = CTYPE_CACHE;
    new->mark    = mark;
    new->flags   = flags;
    new->status  = *status;
    new->used    = 0;

    currCachefiles += 1;
    currCachesize  += new->size;

    insertAtTail ((dll_t *)head, (dll_t *)new);
    if (np != NULL) *np = new;

    return 0;
}

int addACL (dla_t *head, int mark, int perm, char *addr, char *mask, 
            int resolv, dla_t **np)
{
    dla_t *dp, *new, *tail;

    getHeadTail (head, NULL, &tail);

    if ((new = (dla_t *)Malloc(sizeof(dla_t))) == NULL)
        return -1;

    if (mark < 0) {
        dp = (dla_t *)findByMark((dll_t *)head, mark, DLL_GETMAX);
        if (dp == NULL)
            mark = 1;
        else
            mark = dp->mark + 1;
    }

    new->ctype = CTYPE_ACL;
    new->mark = mark;
    new->acl.perm = perm;

    if (!strcasecmp(addr, "host")) {
        new->acl.mask = IPMASK_HOST;
        new->acl.addr = xinet_addr(mask, resolv);
        if (new->acl.addr == INADDR_NONE) {
            Free(new);
            return -2;
        }
    }
    else if (!strcasecmp(addr, "any")) {
        new->acl.mask = IPMASK_ANY;
        new->acl.addr = IPMASK_ANY;
    }
    else {
        new->acl.mask = ((in_addr_t)inet_addr(mask));
        new->acl.addr = xinet_addr(addr, resolv);
        if (new->acl.addr == INADDR_NONE) {
            Free(new);
            return -2;
        }
    }

    for (dp = head->next; dp != tail; dp = dp->next)
        if (dp->mark == mark) {
            if (ntohl(~dp->acl.mask) > ntohl(~new->acl.mask)) 
                break;
        }

    insertBefore((dll_t *)dp, (dll_t *)new);
    if (np != NULL) *np = new;

    return 0;
}

int addTimer (dlt_t *head, int mark, tout_t *tp, dlt_t **np)
{
    dlt_t *dp, *new, *tail;

    if (getTimerByMark(head, NULL, mark) != NULL)
        return -2;

    getHeadTail (head, NULL, &tail);

    if ((new = (dlt_t *)Malloc(sizeof(dlt_t))) == NULL)
        return -1;

    new->ctype = CTYPE_TIMES;
    new->mark  = mark;
    new->times = *tp;

    for (dp = head->next; dp != tail; dp = dp->next)
        if (dp->mark == mark)
            break;

    insertBefore((dll_t *)dp, (dll_t *)new);
    if (np != NULL) *np = new;

    return 0;
}

int addPid (dlp_t *head, int mark, pid_t pid, sai_t *peer, dlp_t **np)
{
    dlp_t *new;

    if ((new = (dlp_t *)Malloc(sizeof(dlp_t))) == NULL)
        return -1;

    new->ctype  = CTYPE_PID;
    new->mark   = mark;
    new->pid    = pid;
    if (peer != NULL)
        new->peer = *peer;

    insertAtTail((dll_t *)head, (dll_t *)new);

    if (np != NULL)
        *np = new;

    return 0;
}

int addBlist (dlb_t *head, int first, int last, dlb_t **np)
{
    dlb_t *new, *tail;

    getHeadTail (head, NULL, &tail);

    if ((new = (dlb_t *)Malloc(sizeof(dlb_t))) == NULL)
        return -1;

    new->ctype  = CTYPE_BLIST;
    new->first = first;
    new->last  = last;

    insertBefore((dll_t *)tail, (dll_t *)new);

    if (np != NULL)
        *np = new;

    return 0;
}

int addFilter (dlf_t *head, int mark, char *rxf, char *txf, dlf_t **np)
{
    dlf_t *new;

    if ((new = (dlf_t *)Malloc(sizeof(dlf_t))) == NULL)
        return -1;

    if ((rxf == NULL) || (*rxf == '\0')) {
        new->rxfilter = NULL;
    }
    else {
        if ((new->rxfilter = Strdup(rxf)) == NULL) {
            Free(new);
            return -1;
        }
    }

    if ((txf == NULL) || (*txf == '\0')) {
        new->txfilter = NULL;
    }
    else {
        if ((new->txfilter = Strdup(txf)) == NULL) {
            if (new->rxfilter != NULL)
                Free(new->rxfilter);
            Free(new);
            return -1;
        }
    }

    new->ctype = CTYPE_FILTER;
    new->mark = mark;

    insertAtTail((dll_t *)head, (dll_t *)new);
    if (np != NULL) *np = new;

    return 0;
}

void sortACLbyIP (dla_t *head)
{
    dla_t *ptr, *next, *tail;
    int ex, pbits, nbits;

    if (head == NULL)
        return;

    getHeadTail (head, NULL, &tail);

    do {
        ex = 0;
        for (ptr = head->next; ptr != tail; ptr = next) {
            if ((next = ptr->next) == tail)
                break;
            pbits = setBits(ptr->acl.mask);
            nbits = setBits(next->acl.mask);
            if (pbits >= nbits) {
                continue;
            }
            else if (pbits == nbits) {
                if (ptr->acl.addr <= next->acl.addr)
                    continue;
            }
            ex++;
            exchangeElements ((dll_t *)ptr, (dll_t *)next);
        }
    } while (ex);
}

void sortFilterByMark (dlf_t *head, int mode)
{
    sortByMark ((dll_t *)head, mode);
}

void sortTimerByMark (dlt_t *head, int mode)
{
    sortByMark ((dll_t *)head, mode);
}

int findNextInBlist (dlb_t *db, int val)
{
    dlb_t *this;
    int new;

    for (new = val+1, this = NULL; (this = getNextBlist(db, this)) != NULL;) {
        if ((new >= this->first) && (new <= this->last))
            return new;
    }

/* we currently cannot do the search in one single loop
   because we cannot be sure the list is sorted ascending */

    for (this = NULL; (this = getNextBlist(db, this)) != NULL;) {
        if (new < this->first)
            return this->first;
    }

   return -1;
}

void dumpBody (FILE *fd, dll_t *dp, int flag, int element) 
{
    acl_t *ap;
    struct in_addr ia;
    static char *cnames[] = { "Unknown", "Plain", "RegEx", "ACL", 
                              "PID", "Filter", "Timer"};
    int csize;

    csize = sizeof(cnames)/sizeof(cnames[0]);
    fprintf(fd, "%c%4d: %-10p  Next: %-10p  Prev: %-10p   Type: %-7s  Mark: %d",
            flag ? '*' : ' ', element, dp, dp->next, dp->prev, 
            (dp->ctype>0 && dp->ctype<csize) ? cnames[dp->ctype] : cnames[0],
            dp->mark);
    switch (dp->ctype) {
        case CTYPE_PLAIN:
            fprintf(fd, "  Data: %d",(int)dp->data[0]);
            break;
        case CTYPE_ACL:
            ap = (acl_t *)dp->data;
            fprintf(fd, "\nPerm: %4.4x",ap->perm);
            ia.s_addr = ap->addr;
            fprintf(fd, "  Addr: %s",inet_ntoa(ia));
            ia.s_addr = ap->mask;
            fprintf(fd, "  Mask: %s",inet_ntoa(ia));
            break;
        case CTYPE_REGEX:
            fprintf(fd, "\nRegex: [%s]",((dlr_t *)dp)->raw);
            break;
        case CTYPE_PID:
            fprintf(fd, "  PID: %d",(int)(((dlp_t *)dp)->pid));
            break;
        case CTYPE_FILTER:
            fprintf(fd, "  Rx-filter: '%s'  Tx-Filter: '%s'",
                    ((dlf_t *)dp)->rxfilter, ((dlf_t *)dp)->txfilter);
            break;
        case CTYPE_TIMES:
            fprintf(fd, "  %s", timer2asc(&((dlt_t *)dp)->times,
                                          T2A_XFER|T2A_SETUP|T2A_RET)); 
            break;
    }
    printf("\n");
}

